@@ -12,6 +12,7 @@ import os
 import pdb
 import random
 import sys
+from StringIO import StringIO  # cStringIO is not pickable
 import time
 import uuid
 
@@ -22,28 +23,31 @@ import yaml
 from pytest_diffeo import redis_address
 import kvlayer
 from kvlayer import BadKey
+from kvlayer._client import STORAGE_CLIENTS
 import yakonfig
 
 logger = logging.getLogger(__name__)
 
-# TODO: fix cassandra for new flexible key-tuples
-#backends = ['local', 'filestorage', 'cassandra', 'accumulo', 'redis']
-backends = ['local', 'filestorage', 'redis', 'accumulo']
-try:
-    import psycopg2
-    backends.append('postgres')
-except ImportError, e:
-    # backends.dont_append('postgres')
-    pass
-
 @pytest.fixture(scope='module',
-                params=backends)
+                params=STORAGE_CLIENTS.keys())
 def backend(request):
-    return request.param
+    backend = request.param
+    if backend == 'cassandra':
+        pytest.skip('cassandra doesn\'t support non-UUID keys')
+    if not request.fspath.dirpath('config_{}.yaml'.format(backend)).exists():
+        pytest.skip('no configuration file for backend {}'.format(backend))
+    return backend
 
 @pytest.yield_fixture(scope='function')
 def client(backend, request, tmpdir, namespace_string):
     config_path = str(request.fspath.dirpath('config_{}.yaml'.format(backend)))
+    # read and parse the config file, insert an object
+    with open(config_path, 'r') as f:
+        file_config = yaml.load(f)
+        # Insert an object into the config which stats will write to.
+        # Below we can get the stats text and log it here.
+        # (Normal stats flow logs to file.)
+        file_config['kvlayer']['log_stats'] = StringIO()
     params = dict(
         app_name='kvlayer',
         namespace=namespace_string,
@@ -58,11 +62,18 @@ def client(backend, request, tmpdir, namespace_string):
     if backend == 'redis':
         params['storage_addresses'] = [ redis_address(request) ]
 
-    with yakonfig.defaulted_config([kvlayer], filename=config_path,
-                                   params=params):
+    with yakonfig.defaulted_config(
+            [kvlayer],
+            config=file_config,
+            params=params):
         client = kvlayer.client()
         client.delete_namespace()
         yield client
+        if client._log_stats is not None:
+            client._log_stats.flush()
+            logger.info('storage stats (%s %s):\n%s',
+                        backend, request.function.__name__,
+                        file_config['kvlayer']['log_stats'].getvalue())
         client.delete_namespace()
 
 def test_basic_storage(client):
@@ -70,7 +81,6 @@ def test_basic_storage(client):
     assert 0 == len(list(client.scan('t1')))
     # use time-based UUID 1, so these are ordered
     u1, u2, u3 = uuid.uuid1(), uuid.uuid1(), uuid.uuid1()
-    #pdb.set_trace()
     client.put('t1', ((u1, u2), b'88'))
     client.put('t2', ((u1, u2, u3), b'88'))
     assert 1 == len(list(client.scan('t1')))
@@ -121,6 +131,13 @@ def test_get(client):
     assert res[0][1] == b'83'
 
 
+def test_put_put(client):
+    client.setup_namespace(dict(t1=1))
+    u1 = uuid.uuid1()
+    client.put('t1', ((u1,),'a'))
+    client.put('t1', ((u1,),'b'))
+    assert list(client.get('t1', (u1,))) == [((u1,),'b')]
+
 
 def test_adding_tables(client):
     client.setup_namespace(dict(t1=2, t2=3))
@@ -174,6 +191,7 @@ def test_setup_namespace_idempotent(client):
     assert 0 == len(list(client.scan('t1', ((u1,), (u1,)))))
 
 
+@pytest.mark.performance
 def test_storage_speed(client):
     client.setup_namespace(dict(t1=2, t2=3))
     num_rows = 10 ** 4
@@ -315,6 +333,84 @@ def test_scan_2d(client):
     assert scan(kf(0,6), kf(6,0)) == kvps
     assert scan(k1f(0), k1f(0)) == []
 
+def test_scan_keys_2d(client):
+    '''Same as test_scan_2d, but only scan the key space.'''
+    client.setup_namespace({'t1': 2})
+    def kf(x, y): return (uuid.UUID(int=x), uuid.UUID(int=y))
+    def k1f(x): return (uuid.UUID(int=x),)
+    def vf(x, y): return '{}:{}'.format(x, y)
+    def kvf(x, y): return (kf(x, y), vf(x, y))
+    def scan_keys(fr, to): return list(client.scan_keys('t1', (fr, to)))
+    r = range(1,5) # [1, 2, 3, 4]
+
+    kvps = [kvf(x, y) for x in r for y in r]
+    keys = [kf(x, y) for x in r for y in r]
+    client.put('t1', *kvps)
+
+    assert list(client.scan_keys('t1')) == keys
+    assert scan_keys(k1f(2), k1f(2)) == [kf(2,y) for y in r]
+    assert scan_keys(kf(2,1), kf(2,2)) == [kf(2,1), kf(2,2)]
+    assert scan_keys(kf(2,0), kf(2,3)) == [kf(2,1), kf(2,2), kf(2,3)]
+    assert scan_keys(kf(2,2), kf(2,5)) == [kf(2,2), kf(2,3), kf(2,4)]
+    assert scan_keys(kf(2,5), kf(3,0)) == []
+    assert scan_keys(kf(2,4), kf(3,1)) == [kf(2,4), kf(3,1)]
+    assert scan_keys(k1f(4), k1f(6)) == [kf(4,y) for y in r]
+    assert scan_keys(k1f(0), k1f(1)) == [kf(1,y) for y in r]
+    assert scan_keys(k1f(0), k1f(6)) == keys
+    assert scan_keys(kf(0,6), kf(6,0)) == keys
+    assert scan_keys(k1f(0), k1f(0)) == []
+
+def test_scan_2d_prefix(client):
+    '''scan(('a',),('a',)) shouldn't find ('ab','c')'''
+    client.setup_namespace({'ts2': (str,str)})
+    client.put('ts2',
+               (('a','a'),'1'),
+               (('a','z'),'2'),
+               (('ab','b'),'3'),
+               (('ab','y'),'4'))
+
+    assert (list(client.scan('ts2', (('a',),('a',)))) ==
+            [(('a','a'),'1'),(('a','z'),'2')])  # but not anything ab
+    assert (list(client.scan_keys('ts2', (('a',),('a',)))) ==
+            [('a','a'),('a','z')])  # but not anything ab
+
+def test_scan_9042(client):
+    tname = 'ts9000'
+    client.setup_namespace({tname: (str,)})
+
+    kv = [
+        (('k{:5x}'.format(x),), 'v{:5x}'.format(x))
+        for x in xrange(9042)
+    ]
+    client.put(tname, *kv)
+
+    # now check that scan gets them all back
+    count = 0
+    for rkv in client.scan(tname):
+        count += 1
+
+    assert count == 9042
+
+    # try other ranged scans
+    count = 0
+    for rkv in client.scan(tname, (('k{:5x}'.format(3000),), ()) ):
+        count += 1
+
+    assert count == 6042
+
+    count = 0
+    for rkv in client.scan(tname, (('k{:5x}'.format(3000),), ('k{:5x}'.format(8437),)) ):
+        count += 1
+
+    assert count == 8437-3000+1 # +1 because scan is inclusive of endpoints
+
+    count = 0
+    for rkv in client.scan(tname, ((), ('k{:5x}'.format(8437),)) ):
+        count += 1
+
+    assert count == 8437+1 # +1 because scan is inclusive of endpoints
+
+
 def test_no_keys(client):
     """Test that standard kvlayer APIs work correctly when not passed keys"""
     client.setup_namespace({'t1': 1})
@@ -375,7 +471,117 @@ def test_new_key_spec(client):
     ]
     client.put('kt4', *good_kt4_kvs)
     count = 0
-    for k,v in client.scan('kt4', (('foo',), ('foo',))):
+    for k,v in client.scan('kt4', (('foo',), ('fooz',))):
         assert (k,v) in good_kt4_kvs
         count += 1
     assert count == len(good_kt4_kvs)
+
+def test_merge_join(client):
+    '''Test that we can read two tables at the same time.'''
+    client.setup_namespace({
+        't1': (str,),
+        't2': (str,),
+    })
+    client.put('t1',
+               (('a',),'one'),
+               (('b',),'two'),
+               (('d',),'four'))
+    client.put('t2',
+               (('a',),'one'),
+               (('c',),'three'))
+    # Find keys in t1 not in t2
+    iter1 = client.scan('t1')
+    iter2 = client.scan('t2')
+    missing = []
+    v1 = next(iter1, None)
+    v2 = next(iter2, None)
+    while v1:
+        if not v2:
+            missing.append(v1)
+            v1 = next(iter1, None)
+        elif v1[0] < v2[0]:
+            missing.append(v1)
+            v1 = next(iter1, None)
+        elif v1[0] == v2[0]:
+            v1 = next(iter1, None)
+            v2 = next(iter2, None)
+        else:
+            v2 = next(iter2, None)
+    assert missing == [(('b',),'two'),(('d',),'four')]
+        
+def test_partial_scan(client):
+    '''Test that reading part of a table, then starting a new scan,
+    doesn't break.'''
+    client.setup_namespace({'t1': (str,)})
+    client.put('t1',
+               (('a',),'one'),
+               (('b',),'two'),
+               (('c',),'three'),
+               (('d',),'four'))
+
+    s1 = client.scan('t1')
+    assert next(s1, None) == (('a',),'one')
+    assert next(s1, None) == (('b',),'two')
+
+    s2 = client.scan('t1')
+    assert next(s2, None) == (('a',),'one')
+    assert next(s2, None) == (('b',),'two')
+    assert next(s2, None) == (('c',),'three')
+    assert next(s2, None) == (('d',),'four')
+    assert next(s2, None) is None
+
+def test_partial_scan_keys(client):
+    '''Test that reading part of a table, then starting a new scan,
+    doesn't break.'''
+    client.setup_namespace({'t1': (str,)})
+    client.put('t1',
+               (('a',),'one'),
+               (('b',),'two'),
+               (('c',),'three'),
+               (('d',),'four'))
+
+    s1 = client.scan('t1')
+    assert next(s1, None) == (('a',),'one')
+    assert next(s1, None) == (('b',),'two')
+
+    s2 = client.scan_keys('t1')
+    assert next(s2, None) == ('a',)
+    assert next(s2, None) == ('b',)
+    assert next(s2, None) == ('c',)
+    assert next(s2, None) == ('d',)
+    assert next(s2, None) is None
+
+def test_scan_name_oddity(client):
+    client.setup_namespace({'index': (str,str,str)})
+    row = (('NAME','alistair','vid'),'1')
+    client.put('index', row)
+    s = client.scan('index', (('NAME','a'), ('NAME','a\xff')))
+    assert list(s) == [row]
+    s = client.scan('index', (('NAME','al'), ('NAME','al\xff')))
+    assert list(s) == [row]
+    s = client.scan('index', (('NAME','ali'), ('NAME','ali\xff')))
+    assert list(s) == [row]
+    s = client.scan('index', (('NAME','alis'), ('NAME','alis\xff')))
+    assert list(s) == [row]
+    s = client.scan('index', (('NAME','alist'), ('NAME','alist\xff')))
+    assert list(s) == [row]
+
+def test_get_returns_keys(client):
+    client.setup_namespace({'t': (str,)})
+    key = ('fubar',)
+    client.put('t', (key, '1'))
+    results = list(client.get('t', key))
+    assert len(results) == 1
+    k, v = results[0]
+    assert k == key
+    assert v == '1'
+
+def test_key_escaping(client):
+    client.setup_namespace({'t': (str, str, str)})
+    key = ('a\0b', 'a%b', 'a%00b')
+    client.put('t', (key, '1'))
+    val = list(client.get('t', key))
+    assert len(val) == 1
+    k, v = val[0]
+    assert k == key
+    assert v == '1'
